# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

from reflekt.property import ReflektProperty


# Traits behave the same way as properties, but traits capture attributes of a
# user, not an event.
class ReflektTrait(ReflektProperty):
    pass



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import copy
import shutil
from pathlib import Path
from typing import Optional, Union

import pkg_resources
import yaml
from inflection import titleize, underscore
from loguru import logger
from packaging.version import Version
from reflekt.columns import reflekt_columns
from reflekt.config import ReflektConfig
from reflekt.dbt import (
    dbt_column_schema,
    dbt_doc_schema,
    dbt_model_schema,
    dbt_src_schema,
    dbt_table_schema,
)
from reflekt.dumper import ReflektYamlDumper
from reflekt.errors import ReflektProjectError
from reflekt.event import ReflektEvent
from reflekt.plan import ReflektPlan
from reflekt.project import ReflektProject
from reflekt.property import ReflektProperty
from reflekt.segment.columns import (
    seg_event_cols,
    seg_groups_cols,
    seg_identify_cols,
    seg_pages_cols,
    seg_screens_cols,
    seg_tracks_cols,
    seg_users_cols,
)
from reflekt.segment.schema import (
    segment_event_schema,
    segment_items_schema,
    segment_payload_schema,
    segment_plan_schema,
    segment_property_schema,
)
from reflekt.trait import ReflektTrait
from reflekt.utils import segment_2_snake
from reflekt.warehouse import WarehouseConnection


class ReflektTransformer(object):
    def __init__(
        self,
        reflekt_plan: ReflektPlan,
        # database: Optional[str] = None,
        # schema: Optional[str] = None,
        dbt_package_name: Optional[str] = None,
        pkg_version: Optional[Version] = None,
    ) -> None:
        self.reflekt_plan = reflekt_plan
        self.plan_name = str.lower(self.reflekt_plan.name)
        self.reflekt_config = ReflektConfig()
        self.cdp_name = self.reflekt_config.cdp_name
        self.plan_type = self.reflekt_config.plan_type
        self.warehouse = self.reflekt_config.warehouse
        self.warehouse_type = self.reflekt_config.warehouse_type
        if reflekt_plan.database is not None:
            self.database = reflekt_plan.database
            self.schema = reflekt_plan.schema
            self.schema_alias = reflekt_plan.schema_alias
            self.reflekt_project = ReflektProject()
            self.project_dir = self.reflekt_project.project_dir
            self.dbt_package_name = dbt_package_name
            self.tmp_pkg_dir = (
                self.project_dir / ".reflekt" / "tmp" / self.dbt_package_name
            )
            self.dbt_pkgs_dir = self.project_dir / "dbt_packages"
            self.dbt_pkg_path = self.dbt_pkgs_dir / self.dbt_package_name
            self.pkg_template = pkg_resources.resource_filename(
                "reflekt", "templates/dbt/"
            )
            self.pkg_version = pkg_version
            self.db_engine = WarehouseConnection()
            self.src_prefix = self.reflekt_project.src_prefix
            self.model_prefix = self.reflekt_project.model_prefix
            self.materialized = self.reflekt_project.materialized
            self.incremental_logic = self.reflekt_project.incremental_logic
            self.docs_prefix = self.reflekt_project.docs_prefix
            self.docs_in_folder = self.reflekt_project.docs_in_folder

    def build_cdp_plan(self, plan_type: Optional[str] = None) -> None:
        if plan_type is None:
            plan_type = self.plan_type

        if plan_type.lower() == "rudderstack":
            return self._build_plan_rudderstack(self.reflekt_plan)
        elif plan_type.lower() == "segment":
            return self._build_plan_segment(self.reflekt_plan)
        elif plan_type.lower() == "snowplow":
            return self._build_plan_snowplow(self.reflekt_plan)
        # 'reflekt push' not supported for Avo or Iteratively.

    def _build_plan_rudderstack(self):
        pass

    def _build_plan_snowplow(self):
        pass

    def _build_plan_segment(self, reflekt_plan: ReflektPlan) -> dict:
        segment_payload = copy.deepcopy(segment_payload_schema)
        segment_plan = copy.deepcopy(segment_plan_schema)
        segment_plan["display_name"] = self.plan_name
        print("")  # Terminal newline
        logger.info(
            f"Converting Reflekt tracking plan '{self.plan_name}' to "
            f"{titleize(self.plan_type)} format"
        )

        if reflekt_plan.events != []:
            for reflekt_event in reflekt_plan.events:
                segment_event = self._build_segment_event(reflekt_event)
                segment_plan["rules"]["events"].append(segment_event)

        if reflekt_plan.identify_traits != []:
            for reflekt_identify_trait in reflekt_plan.identify_traits:
                segment_identify_trait = self._build_segment_trait(
                    reflekt_identify_trait
                )

                if reflekt_identify_trait.required:
                    segment_plan["rules"]["identify"]["properties"]["traits"][
                        "required"
                    ].append(reflekt_identify_trait.name)

                segment_plan["rules"]["identify"]["properties"]["traits"][
                    "properties"
                ].update({reflekt_identify_trait.name: segment_identify_trait})

        if reflekt_plan.group_traits != []:
            for reflekt_group_trait in reflekt_plan.group_traits:
                segment_group_trait = self._build_segment_trait(reflekt_group_trait)

                if reflekt_group_trait.required:
                    segment_plan["rules"]["group"]["properties"]["traits"][
                        "required"
                    ].append(reflekt_group_trait.name)

                segment_plan["rules"]["group"]["properties"]["traits"][
                    "properties"
                ].update({reflekt_group_trait.name: segment_group_trait})

        segment_payload["tracking_plan"].update(segment_plan)

        return segment_payload

    def _build_segment_event(self, reflekt_event: ReflektEvent) -> dict:
        if reflekt_event.version > 1:
            version_str = f" (version {reflekt_event.version})"
        else:
            version_str = ""

        logger.info(f"    Converting '{reflekt_event.name}'{version_str}")
        segment_event = copy.deepcopy(segment_event_schema)
        segment_event["name"] = reflekt_event.name
        segment_event["description"] = reflekt_event.description
        segment_event["version"] = reflekt_event.version
        segment_event["rules"]["labels"] = reflekt_event.metadata

        for reflekt_property in reflekt_event.properties:
            segment_property = self._build_segment_property(reflekt_property)

            if reflekt_property.required:
                segment_event["rules"]["properties"]["properties"]["required"].append(
                    reflekt_property.name
                )

            segment_event["rules"]["properties"]["properties"]["properties"].update(
                {reflekt_property.name: segment_property}
            )

        return segment_event

    def _build_segment_property(self, reflekt_property: ReflektProperty) -> dict:
        segment_property = copy.deepcopy(segment_property_schema)
        segment_property["description"] = reflekt_property.description
        segment_property["type"] = [reflekt_property.type]
        segment_property = self._parse_reflekt_property(
            reflekt_property, segment_property
        )

        return segment_property

    def _parse_reflekt_property(
        self,
        reflekt_property: Union[ReflektProperty, ReflektTrait],
        segment_property: dict,
    ) -> dict:
        # NOTE - ReflektTrait is a child class of ReflektProperty, so this
        # instance method will parse Reflekt properties and traits
        updated_segment_property = copy.deepcopy(segment_property)
        if hasattr(reflekt_property, "allow_null") and reflekt_property.allow_null:
            updated_segment_property["type"].append("null")
        elif hasattr(reflekt_property, "type") and reflekt_property.type == "any":
            # If data type is 'any', delete type key. Otherwise, triggers
            # 'invalid argument' from Segment API
            updated_segment_property.pop("type", None)
        elif (
            hasattr(reflekt_property, "pattern") and reflekt_property.pattern is not None
        ):
            updated_segment_property["pattern"] = reflekt_property.pattern
        elif hasattr(reflekt_property, "enum") and reflekt_property.enum is not None:
            updated_segment_property["enum"] = reflekt_property.enum
        elif hasattr(reflekt_property, "datetime") and reflekt_property.datetime:
            updated_segment_property["format"] = "date-time"

        if (
            hasattr(reflekt_property, "array_item_schema")
            and reflekt_property.array_item_schema is not None
        ):
            segment_array_items = copy.deepcopy(segment_items_schema)
            for reflekt_item_property in reflekt_property.array_item_schema:
                segment_item_property = copy.deepcopy(segment_property_schema)
                segment_item_property["description"] = reflekt_item_property[
                    "description"
                ]
                segment_item_property["type"] = [reflekt_item_property["type"]]
                segment_item_property = self._parse_reflekt_property(
                    reflekt_item_property, segment_item_property
                )

                if (
                    "required" in reflekt_item_property
                    and reflekt_item_property["required"]
                ):
                    segment_array_items["items"]["required"].append(
                        reflekt_item_property["name"]
                    )

                segment_array_items["items"]["properties"].update(
                    {reflekt_item_property["name"]: segment_item_property}
                )

            updated_segment_property.update(segment_array_items)

        if (
            hasattr(reflekt_property, "object_properties")
            and reflekt_property.object_properties is not None
        ):
            updated_segment_property["properties"] = {}
            updated_segment_property["required"] = []

            for object_property in reflekt_property.object_properties:
                segment_property = copy.deepcopy(segment_property_schema)
                segment_property["description"] = object_property["description"]
                segment_property["type"] = object_property["type"]
                updated_segment_property["properties"].update(
                    {object_property["name"]: segment_property}
                )

                if "required" in object_property and object_property["required"]:
                    updated_segment_property["required"].append(object_property["name"])

        return updated_segment_property

    def _build_segment_trait(self, reflekt_trait: ReflektTrait) -> dict:
        segment_trait = copy.deepcopy(segment_property_schema)
        segment_trait["description"] = reflekt_trait.description
        segment_trait["type"] = [reflekt_trait.type]
        # _parse_reflekt_property works on traits too
        segment_trait = self._parse_reflekt_property(reflekt_trait, segment_trait)

        return segment_trait

    def check_for_schema_match(self) -> None:
        pass

    def build_dbt_package(self) -> None:
        logger.info(
            f"Building Reflekt dbt package:"
            f"\n        cdp: {self.cdp_name}"
            f"\n        analytics governance tool: {self.plan_type}"
            f"\n        tracking plan: {self.plan_name}"
            f"\n        warehouse: {self.warehouse_type}"
            f"\n        schema: {self.schema}"
            f"\n        schema_alias: {self.schema_alias}"
            f"\n        dbt package name: {self.dbt_package_name}"
            f"\n        dbt package version: {self.pkg_version}"
            f"\n        dbt package path: {self.dbt_pkg_path}\n"
        )

        if self.tmp_pkg_dir.exists():
            shutil.rmtree(str(self.tmp_pkg_dir))
        shutil.copytree(self.pkg_template, str(self.tmp_pkg_dir))

        # Update the version string in templated dbt_project.yml
        with open(self.tmp_pkg_dir / "dbt_project.yml", "r") as f:
            dbt_project_yml_str = f.read()

        dbt_project_yml_str = dbt_project_yml_str.replace(
            "0.1.0", str(self.pkg_version)  # Template always has version '0.1.0'
        ).replace("reflekt_package_name", self.dbt_package_name)

        with open(self.tmp_pkg_dir / "dbt_project.yml", "w") as f:
            f.write(dbt_project_yml_str)

        # Set dbt_pkg_name and plan_name in README.md
        with open(self.tmp_pkg_dir / "README.md", "r") as f:
            readme_str = f.read()

        readme_str = readme_str.replace("_DBT_PKG_NAME_", self.dbt_package_name).replace(
            "_PLAN_NAME_", self.plan_name
        )

        with open(self.tmp_pkg_dir / "README.md", "w") as f:
            f.write(readme_str)

        if self.cdp_name == "rudderstack":
            return self._dbt_package_rudderstack(self.reflekt_plan)
        elif self.cdp_name == "segment":
            return self._dbt_package_segment(self.reflekt_plan)
        elif self.cdp_name == "snowplow":
            return self._dbt_package_snowplow(self.reflekt_plan)

    def _dbt_package_rudderstack(self):
        pass

    def _dbt_package_snowplow(self):
        pass

    def _dbt_package_segment(self, reflekt_plan: ReflektPlan) -> None:
        self.db_errors = []

        # Setup `Page Viewed` and `Screen Viewed` events, if they exist in Reflekt plan
        page_viewed_event = [
            event
            for event in reflekt_plan.events
            if str.lower(event.name).replace("_", " ").replace("-", " ") == "page viewed"
        ]

        if page_viewed_event == []:  # Page Viewed not in tracking plan
            page_viewed_props = []
        else:
            page_viewed_props = page_viewed_event[0].properties

        screen_viewed_event = [
            event
            for event in reflekt_plan.events
            if str.lower(event.name).replace("_", " ").replace("-", " ")
            == "screen viewed"
        ]

        if screen_viewed_event == []:  # Screen Viewed not in tracking plan
            screen_viewed_props = []
        else:
            screen_viewed_props = screen_viewed_event[0].properties

        std_segment_tables = [
            {
                "name": "identifies",
                "description": "A table with identify() calls.",  # noqa: E501
                "unique_key": "identify_id",
                "cdp_cols": seg_identify_cols,
                "plan_cols": [],
            },
            {
                "name": "users",
                "description": "A table with the latest traits for users.",  # noqa: E501
                "unique_key": "user_id",
                "cdp_cols": seg_users_cols,
                "plan_cols": [],
            },
            {
                "name": "groups",
                "description": "A table with group() calls.",  # noqa: E501
                "unique_key": "group_id",
                "cdp_cols": seg_groups_cols,
                "plan_cols": [],
            },
            {
                "name": "pages",
                "description": "A table with page() calls.",  # noqa: E501
                "unique_key": "page_id",
                "cdp_cols": seg_pages_cols,
                "plan_cols": page_viewed_props,
            },
            {
                "name": "screens",
                "description": "A table with screen() calls.",  # noqa: E501
                "unique_key": "screen_id",
                "cdp_cols": seg_screens_cols,
                "plan_cols": screen_viewed_props,
            },
            {
                "name": "tracks",
                "description": "A table with track() event calls.",  # noqa: E501
                "unique_key": "event_id",
                "cdp_cols": seg_tracks_cols,
                "plan_cols": [],
            },
        ]

        dbt_src = self._template_dbt_source(reflekt_plan=reflekt_plan)

        for std_segment_table in std_segment_tables:
            if self.plan_type == "avo" and std_segment_table["name"] in [
                "identifies",
                "users",
                "groups",
            ]:
                pass
            else:
                db_columns, error_msg = self.db_engine.get_columns(
                    self.schema, std_segment_table["name"]
                )
                if error_msg is not None:
                    self.db_errors.append(error_msg)
                else:
                    table_name = std_segment_table["name"]
                    model_name = (
                        f"{self.model_prefix}{self.schema}__{table_name}"
                        if self.schema_alias is None
                        else f"{self.model_prefix}{self.schema_alias}__{table_name}"
                    )
                    doc_name = (
                        f"{self.docs_prefix}{self.schema}__{table_name}"
                        if self.schema_alias is None
                        else f"{self.docs_prefix}{self.schema_alias}__{table_name}"
                    )
                    self._template_dbt_table(
                        dbt_src=dbt_src,
                        table_name=table_name,
                        table_description=std_segment_table["description"],
                    )
                    self._template_dbt_model(
                        materialized=self.materialized,
                        unique_key=std_segment_table["unique_key"],
                        table_name=table_name,
                        model_name=model_name,
                        db_columns=db_columns,
                        cdp_cols=std_segment_table["cdp_cols"],
                        plan_cols=std_segment_table["plan_cols"],
                    )
                    self._template_dbt_doc(
                        doc_name=doc_name,
                        model_name=model_name,
                        model_description=std_segment_table["description"],
                        db_columns=db_columns,
                        cdp_cols=std_segment_table["cdp_cols"],
                        plan_cols=std_segment_table["plan_cols"],
                    )

        for event in reflekt_plan.events:
            if event.name.lower() in ["page viewed", "screen viewed"]:
                pass  # Already handled above by std_segment_tables for loop
            else:
                db_columns, error_msg = self.db_engine.get_columns(
                    self.schema, segment_2_snake(event.name)
                )
                if error_msg is not None:
                    self.db_errors.append(error_msg)
                else:
                    table_name = segment_2_snake(event.name)
                    model_name = (
                        f"{self.model_prefix}{self.schema}__{table_name}"
                        if self.schema_alias is None
                        else f"{self.model_prefix}{self.schema_alias}__{table_name}"
                    )
                    doc_name = (
                        f"{self.docs_prefix}{self.schema}__{table_name}"
                        if self.schema_alias is None
                        else f"{self.docs_prefix}{self.schema_alias}__{table_name}"
                    )
                    self._template_dbt_table(
                        dbt_src=dbt_src,
                        table_name=table_name,
                        table_description=event.description,
                    )
                    self._template_dbt_model(
                        materialized=self.materialized,
                        unique_key="event_id",
                        table_name=table_name,
                        model_name=model_name,
                        db_columns=db_columns,
                        cdp_cols=seg_event_cols,
                        plan_cols=event.properties,
                    )
                    self._template_dbt_doc(
                        doc_name=doc_name,
                        model_name=model_name,
                        model_description=event.description,
                        db_columns=db_columns,
                        cdp_cols=seg_event_cols,
                        plan_cols=event.properties,
                    )

        src_name = (
            f"{self.src_prefix}{self.schema}"
            if self.schema_alias is None
            else f"{self.src_prefix}{self.schema_alias}"
        )
        dbt_src_path = self.tmp_pkg_dir / "models" / f"{src_name}.yml"

        if self.docs_in_folder:
            docs_folder_str = str(self.tmp_pkg_dir) + "/models/docs"
            docs_folder_path = Path(docs_folder_str)
            docs_folder_path.mkdir(exist_ok=True)
            dbt_src_path = docs_folder_path / f"{src_name}.yml"
        else:
            dbt_src_path = self.tmp_pkg_dir / "models" / f"{src_name}.yml"

        with open(dbt_src_path, "w") as f:
            yaml.dump(
                dbt_src,
                f,
                indent=2,
                width=70,
                Dumper=ReflektYamlDumper,
                sort_keys=False,
                default_flow_style=False,
                allow_unicode=True,
                encoding=("utf-8"),
            )

        db_errors_list = ["    " + error + "\n" for error in self.db_errors]
        db_errors_str = ""

        for db_error in db_errors_list:
            db_errors_str += db_error

        if self.db_errors:
            print("")  # Terminal newline
            logger.warning(
                f"[WARNING] The following database error(s) were encountered "
                f"while templating the dbt package.\n"
                f"  NOTE - these errors do not prevent templating.\n\n{db_errors_str}"
            )

        logger.info(
            f"Copying dbt package from temporary path "
            f"{self.tmp_pkg_dir} to {self.dbt_pkg_path}"
        )

        if self.dbt_pkg_path.exists():
            shutil.rmtree(self.dbt_pkg_path)

        shutil.copytree(self.tmp_pkg_dir, self.dbt_pkg_path)

        print("")  # Terminal newline
        logger.info(
            f"[SUCCESS] Created Reflekt dbt package '{self.dbt_package_name}' "
            f"at: {self.dbt_pkg_path}"
        )
        print("")  # Cleanup stdout

    def _template_dbt_source(self, reflekt_plan: ReflektPlan) -> dict:
        logger.info(f"Initializing template for dbt source {self.schema}")
        dbt_src = copy.deepcopy(dbt_src_schema)
        dbt_src["sources"][0]["name"] = (
            self.schema_alias if self.schema_alias is not None else self.schema
        )
        dbt_src["sources"][0]["schema"] = self.schema
        dbt_src["sources"][0]["database"] = self.database
        dbt_src["sources"][0]["description"] = (
            f"Schema in {titleize(self.warehouse_type)} where data for the "
            f"{reflekt_plan.name} {titleize(self.cdp_name)} source is stored."
        )

        return dbt_src

    def _template_dbt_table(
        self,
        dbt_src: dict,
        table_name: str,
        table_description: str,
        # db_columns: list,
        # cdp_cols: dict,
        # plan_cols: list,
    ) -> None:
        print("")  # Terminal newline
        logger.info(f"Templating table '{table_name}' in dbt source {self.schema}")
        dbt_tbl = copy.deepcopy(dbt_table_schema)
        dbt_tbl["name"] = table_name
        dbt_tbl["description"] = table_description
        dbt_src["sources"][0]["tables"].append(dbt_tbl)

    def _template_dbt_model(
        self,
        materialized: str,
        unique_key: str,
        table_name: str,
        model_name: str,
        db_columns: list,
        cdp_cols: dict,
        plan_cols: list,
    ) -> None:
        print("")  # Terminal newline
        logger.info(
            f"Templating dbt model "
            f"{self.model_prefix}{self.schema}__{table_name}.sql"
        )
        logger.info("    Adding {{ config(...) }} to model SQL")
        mdl_sql = self._template_dbt_model_config(
            materialized,
            unique_key,
            self.warehouse_type,
        )
        logger.info("    Adding source CTE to model SQL")
        source_schema = (
            self.schema_alias if self.schema_alias is not None else self.schema
        )
        mdl_sql += self._template_dbt_source_cte(
            source_schema=source_schema,
            source_table=table_name,
            incremental_logic=self.incremental_logic,
        )

        logger.info("    Adding renamed CTE to model SQL")
        mdl_sql += "renamed as (\n\n" "    select"

        for column, mapped_columns in cdp_cols.items():
            if column in db_columns or column in reflekt_columns:
                for mapped_column in mapped_columns:
                    if mapped_column["schema_name"] is not None:
                        logger.info(
                            f"    Adding column '{mapped_column['schema_name']}' to "
                            f"model SQL"
                        )
                        col_sql = (
                            mapped_column["sql"]
                            .replace("__SCHEMA_NAME__", self.schema)
                            .replace("__TABLE_NAME__", table_name)
                            .replace("__PLAN_NAME__", self.plan_name)
                        )
                        mdl_sql += "\n        " + col_sql + ","

        for column in plan_cols:
            logger.info(
                f"    Adding column '{segment_2_snake(column.name)}' to model SQL"
            )
            col_sql = segment_2_snake(column.name)
            mdl_sql += "\n        " + col_sql + ","

        mdl_sql = mdl_sql[:-1]  # Remove final trailing comma
        # fmt: off
        mdl_sql += (
            "\n\n    from source"
            "\n\n)"
            "\n\n"
            "select * from renamed\n"
        )
        # fmt: on
        model_path = self.tmp_pkg_dir / "models" / f"{model_name}.sql"

        with open(model_path, "w") as f:
            f.write(mdl_sql)

    def _template_dbt_model_config(
        self,
        materialized: str,
        unique_key: str,
        warehouse_type: str,
    ) -> str:
        if materialized == "view":
            # fmt: off
            model_config = (
                "{{\n"
                "  config(\n"
                "    materialized = 'view',\n"
                "  )\n"
                "}}\n\n"
            )
            # fmt: on
        elif materialized == "incremental":
            if warehouse_type == "redshift":
                sort_str = "sort = 'tstamp'"
            elif warehouse_type == "snowflake":
                sort_str = "cluster_by = 'tstamp'"
            model_config = (
                "{{\n"
                f"  config(\n"
                f"    materialized = 'incremental',\n"
                f"    unique_key = '{unique_key}',\n"
                f"    {sort_str}\n"
                f"  )\n"
                "}}\n\n"
            )
        else:
            raise ReflektProjectError(
                "Invalid materialized config in reflekt_project.yml. Must be "
                "either 'view' or 'incremental'. See Reflekt docs on project "
                "configuration:"
                "    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

        return model_config

    def _template_dbt_source_cte(
        self,
        source_schema: str,
        source_table: str,
        incremental_logic: Optional[str] = None,
    ) -> str:
        if incremental_logic is None:
            incremental_logic = ""

        # Format incremental_logic from reflekt_project.yml so it templates
        # according to dbt-labs style guide https://bit.ly/383kG7l
        incremental_logic_list = incremental_logic.splitlines()
        incremental_logic_str = ""
        for line in incremental_logic_list:
            incremental_logic_str += f"    {line}\n"

        source_cte = (
            "with\n\n"
            "source as (\n\n"
            f"    select *\n\n"
            f"    from {{{{ source('{underscore(source_schema)}', '{source_table}') }}}}\n"  # noqa: E501
            f"{incremental_logic_str}\n"
            "),\n\n"
        )

        return source_cte

    def _template_dbt_doc(
        self,
        doc_name: str,
        model_name: str,
        model_description: str,
        db_columns: list,
        cdp_cols: dict,
        plan_cols: list,
    ) -> None:
        print("")  # Terminal newline
        logger.info(
            f"Templating dbt docs " f"{doc_name}.yml" f" for model " f"{model_name}.sql"
        )
        dbt_doc = copy.deepcopy(dbt_doc_schema)
        dbt_mdl_doc = copy.deepcopy(dbt_model_schema)
        dbt_mdl_doc["name"] = model_name
        dbt_mdl_doc["description"] = model_description

        for column, mapped_columns in cdp_cols.items():
            if column in db_columns or column in reflekt_columns:
                for mapped_column in mapped_columns:
                    if mapped_column["schema_name"] is not None:
                        logger.info(
                            f"    Adding column '{mapped_column['schema_name']}' to "
                            f"dbt docs"
                        )
                        mdl_col = copy.deepcopy(dbt_column_schema)
                        mdl_col["name"] = mapped_column["schema_name"]
                        mdl_col["description"] = mapped_column["description"]
                        if mapped_column.get("tests") is not None:
                            mdl_col["tests"] = mapped_column["tests"]
                        dbt_mdl_doc["columns"].append(mdl_col)

        for column in plan_cols:
            logger.info(f"    Adding column '{segment_2_snake(column.name)}' to docs")
            mdl_col = copy.deepcopy(dbt_column_schema)
            mdl_col["name"] = segment_2_snake(column.name)
            mdl_col["description"] = column.description
            dbt_mdl_doc["columns"].append(mdl_col)

        dbt_doc["models"].append(dbt_mdl_doc)

        if self.docs_in_folder:
            docs_folder_str = str(self.tmp_pkg_dir) + "/models/docs"
            docs_folder_path = Path(docs_folder_str)
            docs_folder_path.mkdir(exist_ok=True)
            docs_path = docs_folder_path / f"{doc_name}.yml"
        else:
            docs_path = self.tmp_pkg_dir / "models" / f"{doc_name}.yml"

        with open(docs_path, "w") as f:
            yaml.dump(
                dbt_doc,
                f,
                indent=2,
                width=70,
                Dumper=ReflektYamlDumper,
                sort_keys=False,
                default_flow_style=False,
                allow_unicode=True,
                encoding=("utf-8"),
            )



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
# SPDX-License-Identifier: Apache-2.0
#
# SPDX-FileCopyrightText: 2021 Buffer
# SPDX-License-Identifier: MIT

import re
from collections import Counter

from cerberus import Validator

from reflekt.casing import CAMEL_CASE_RE, SNAKE_CASE_RE, TITLE_CASE_RE
from reflekt.errors import ReflektValidationError
from reflekt.project import ReflektProject
from reflekt.property import ReflektProperty
from reflekt.schema import reflekt_event_schema, reflekt_expected_metadata_schema


# The class ReflektEvent is a derivative work based on the class
# YamlEvent from project tracking-plan-kit licensed under MIT. All
# changes are licensed under Apache-2.0.
class ReflektEvent(object):
    def __init__(self, event_yaml_obj: dict) -> None:
        if ReflektProject().exists:
            self._project = ReflektProject()
            self._event_yaml_obj = event_yaml_obj
            self._properties = [
                ReflektProperty(property)
                for property in self._event_yaml_obj["properties"]
            ]
            self.version = self._event_yaml_obj.get("version")
            self.name = self._event_yaml_obj.get("name")
            self.description = self._event_yaml_obj.get("description")
            self.metadata = self._event_yaml_obj.get("metadata")
            self.properties = [
                ReflektProperty(property)
                for property in self._event_yaml_obj["properties"]
            ]
            self.validate_event()

    def _check_event_metadata(self) -> None:
        if self.metadata:
            validator = Validator(reflekt_expected_metadata_schema)
            is_valid = validator.validate(
                self.metadata, reflekt_expected_metadata_schema
            )
            if not is_valid:
                raise ReflektValidationError(
                    f"Invalid metadata specified for event '{self.name} - "
                    f"{validator.errors}\n\n"
                    f"See Reflekt docs on defining metadata and metadata tests:"
                    f"\n    Event metadata: https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#metadata"  # noqa: E501
                    f"\n    Defining metadata tests: https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
                )

    def _check_event_name_case(self) -> None:
        case_rule = self._project.events_case

        if case_rule is not None:
            rule_str = f"case: {case_rule.lower()}"
            if case_rule.lower() == "title":
                regex = TITLE_CASE_RE
            elif case_rule.lower() == "snake":
                regex = SNAKE_CASE_RE
            elif case_rule.lower() == "camel":
                regex = CAMEL_CASE_RE

            match = bool(re.match(regex, self.name))
            rule_type = "case:"
            rule_str = case_rule.lower()

            if not match:
                raise ReflektValidationError(
                    f"Event name '{self.name}' does not match naming convention"
                    f" defined by '{rule_type} {rule_str}' in reflekt_project.yml. "
                    f"See Reflekt docs on defining naming conventions:\n"
                    f"    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#project-configuration"  # noqa: E501
                )

    def _check_event_name_numbers(self) -> None:
        allow_numbers = self._project.events_allow_numbers
        if not allow_numbers:
            contains_number = any(char.isdigit() for char in self.name)

            if contains_number:
                raise ReflektValidationError(
                    f"\nEvent name '{self.name}' does not match naming convention"
                    f" defined by 'allow_numbers: {str(allow_numbers).lower()}' in reflekt_project.yml "  # noqa: E501
                    f"See Reflekt docs on defining naming conventions:\n"
                    f"    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#project-configuration"  # noqa: E501
                )

    def _check_duplicate_properties(self) -> None:
        if len(self.properties) == 0:
            return

        prop_names = [p.name for p in self.properties]
        counts = Counter(prop_names)

        duplicates = {k: v for (k, v) in counts.items() if v > 1}
        if len(duplicates) > 0:
            duplicate_names = ", ".join(duplicates.keys())
            raise ReflektValidationError(
                f"Duplicate properties found on event {self.name}. "
                f"Properties: {duplicate_names}"
            )

    def _check_reserved_property_names(self) -> None:
        if len(self.properties) == 0:
            return

        prop_names = [p.name for p in self.properties]

        for prop_name in prop_names:
            if prop_name in ReflektProject().properties_reserved:
                raise ReflektValidationError(
                    f"Property name '{prop_name}' is reserved and cannot be used."
                    f"See Reflekt docs on defining reserved property names:\n"
                    f"    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#project-configuration"  # noqa: E501
                )

    def validate_event(self) -> None:
        """Validate event against Reflekt schema."""
        validator = Validator(reflekt_event_schema)
        is_valid = validator.validate(self._event_yaml_obj, reflekt_event_schema)

        if not is_valid:
            raise ReflektValidationError(
                f"Event validation error for event '{self.name}' - {validator.errors}"  # noqa: E501
                f"\n\nSee Reflekt docs on event definition:\n"
                f"    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#events"  # noqa: E501
            )

        self._check_event_name_case()
        self._check_event_name_numbers()
        self._check_duplicate_properties()
        self._check_reserved_property_names()

        if reflekt_expected_metadata_schema is not None:
            self._check_event_metadata()



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

# fmt: off
PLANS = [
    "segment",
    "avo",
    # "rudderstack",
    # "snowplow",
    # "iteratively",
]

PLAN_INIT_STRING = (
    "\n[1] Segment Protocols"
    "\n[2] Avo"
    # "\n[3] Rudderstack"
    # "\n[4] Snowplow"
    # "\n[5] Iteratively"
)

PLAN_MAP = {
    "1": "segment",
    "2": "avo",
    # "3": "rudderstack",
    # "4": "snowplow",
    # "5": "iteratively",
}

CDPS = [
    "segment",
    # "rudderstack",
    # "snowplow",
]

CDP_INIT_STRING = (
    "\n[1] Segment"
    # "\n[2] Rudderstack"
    # "\n[3] Snowplow"
)

CDP_MAP = {
    "1": "segment",
    # "2": "rudderstack",
    # "3": "snowplow",
}

WAREHOUSES = [
    "snowflake",
    "redshift",
]

WAREHOUSE_INIT_STRING = (
    "\n[1] Snowflake"
    "\n[2] Redshift"
)

WAREHOUSE_MAP = {
    "1": "snowflake",
    "2": "redshift",
}
# fmt: on



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import re


def segment_2_snake(in_str: str) -> str:
    """Converts a planCase/PlanCase/PLanCase string to snake_case. Inserts
    underscore only where casing switches from lower to UPPER, lower() entire
    string, replace period with with underscore, replace space with underscore.

    Examples:
        planCase --> plan_case
        PlanCase --> plan_case
        PLanCase --> plan_case

    This function is useful when dealing with column names from Segment
    tracking plans.
    """
    temp = re.sub("([a-z])([A-Z]+)", r"\1_\2", in_str)
    temp2 = temp.lower()
    warehouse_case = (
        temp2.replace(".", "_").replace(" ", "_").replace("_-_", "_").replace("/", "_")
    )
    return warehouse_case



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
import shutil
import subprocess
from pathlib import Path
from typing import Optional

import click
import pkg_resources
import yaml
from inflection import titleize
from loguru import logger
from packaging.version import InvalidVersion
from packaging.version import parse as parse_version

from reflekt import constants
from reflekt.api_handler import ReflektApiHandler
from reflekt.avo.plan import AvoPlan
from reflekt.config import ReflektConfig
from reflekt.loader import ReflektLoader
from reflekt.logger import logger_config
from reflekt.project import ReflektProject
from reflekt.segment.plan import SegmentPlan
from reflekt.transformer import ReflektTransformer


@click.group()
def cli():
    logger.configure(**logger_config)


@click.option(
    "--project-dir",
    "project_dir_str",
    # default=".",
    required=True,
    help="Path where Reflekt project will be created. Defaults to current directory.",
)
@click.command()
def init(project_dir_str: str) -> None:
    """Create a Reflekt project at the provide directory."""
    project_dir = Path(project_dir_str).resolve()
    project_name = click.prompt(
        "Enter your project name (letters, digits, underscore)", type=str
    )
    project_yml = project_dir / "reflekt_project.yml"
    readme_md = project_dir / "README.md"

    if project_yml.exists():
        with open(project_yml, "r") as f:
            reflekt_project_obj = yaml.safe_load(f)

        if reflekt_project_obj["name"] == project_name:
            logger.error(
                f"A Reflekt project named {project_name} already exists in "
                f"{Path.cwd()}"
            )
            raise click.Abort()

    reflekt_config_dir = click.prompt(
        "Enter path to the DIRECTORY where Reflekt will write your "
        "reflekt_config.yml to be use by this Reflekt project",
        type=str,
        default=str(Path.home() / ".reflekt"),
    )
    reflekt_config_dir = Path(reflekt_config_dir)
    reflekt_config_path = reflekt_config_dir / "reflekt_config.yml"
    reflekt_config_path = Path(reflekt_config_path)
    collect_config = True  # Default to collect config from user
    config_name = click.prompt(
        "Enter a config profile name for the reflekt_config.yml", type=str
    )

    if not reflekt_config_path.exists():
        if not reflekt_config_dir.exists():
            reflekt_config_dir.mkdir(parents=True)
        reflekt_config_path.touch()
        reflekt_config_obj = {
            config_name: {
                "plan_type": "",
                "cdp": "",
                "warehouse": {},
            }
        }
    else:
        with open(reflekt_config_path, "r") as f:
            reflekt_config_obj = yaml.safe_load(f)

        if config_name in reflekt_config_obj:
            logger.error(
                f"A Reflekt config profile named {config_name} already exists in "
                f"{reflekt_config_path}"
            )
            raise click.Abort()
        else:
            reflekt_config_obj.update(
                {
                    config_name: {
                        "plan_type": "",
                        "cdp": "",
                        "warehouse": {},
                    }
                }
            )

    if collect_config:
        plan_type_prompt = click.prompt(
            f"What Analytics Governance tool do you use to manage your tracking plan(s)?"
            f"{constants.PLAN_INIT_STRING}"
            f"\nEnter a number",
            type=int,
        )
        plan_type = constants.PLAN_MAP[str(plan_type_prompt)]
        reflekt_config_obj[config_name]["plan_type"] = plan_type

        if plan_type == "segment":
            # workspace_name and access_token required for Segment API use
            workspace_name = click.prompt(
                "Enter your Segment workspace name. You can find this in your Segment "
                "account URL (i.e. https://app.segment.com/your-workspace-name/)",
                type=str,
            )
            reflekt_config_obj[config_name]["workspace_name"] = workspace_name
            access_token = click.prompt(
                "Enter your Segment Config API access token (To generate a "
                "token, see Segment's documentation https://segment.com/docs/config-api/authentication/#create-an-access-token)",  # noqa: E501
                type=str,
                hide_input=True,
            )
            reflekt_config_obj[config_name]["access_token"] = access_token
        elif plan_type == "avo":
            avo_end_msg = (
                "You've selected Avo as your Analytics Governance tool which requires "
                "additional setup steps. Please see the docs for additional guidance:\n"
                "    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md/#connect-reflekt--avo"  # noqa: E501
            )

        # TODO - Enable support for other CDPs below as developed
        # elif plan_type == "rudderstack":
        #     pass
        # elif plan_type == "snowplow":
        #     pass
        # elif plan_type == "iteratively":
        #     pass

        cdp_num_prompt = click.prompt(
            f"How do you collect first-party data?"
            f"{constants.CDP_INIT_STRING}"
            f"\nEnter a number",
            type=int,
        )
        cdp = constants.CDP_MAP[str(cdp_num_prompt)]
        reflekt_config_obj[config_name]["cdp"] = cdp
        warehouse_num = click.prompt(
            f"Which data warehouse do you use?"
            f"{constants.WAREHOUSE_INIT_STRING}"
            f"\nEnter a number",
            type=int,
        )
        warehouse = constants.WAREHOUSE_MAP[str(warehouse_num)]
        reflekt_config_obj[config_name]["warehouse"].update({warehouse: {}})

        if warehouse == "snowflake":
            account = click.prompt("account", type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"account": account}
            )
            user = click.prompt("user", type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"user": user}
            )
            password = click.prompt("password", hide_input=True, type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"password": password}
            )
            role = click.prompt("role", type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"role": role}
            )
            database = click.prompt(
                "database (where raw event data is stored)", type=str
            )
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"database": database}
            )
            snowflake_warehouse = click.prompt("warehouse", type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"warehouse": snowflake_warehouse}
            )
        elif warehouse == "redshift":
            host_url = click.prompt(
                "host_url (hostname.region.redshift.amazonaws.com)", type=str
            )
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"host_url": host_url}
            )
            port = click.prompt("port", default=5439, type=int)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"port": port}
            )
            user = click.prompt("user", type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"user": user}
            )
            password = click.prompt("password", hide_input=True, type=str)
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"password": password}
            )
            db_name = click.prompt(
                "db_name (database where raw event data is stored)",
                type=str,
            )
            reflekt_config_obj[config_name]["warehouse"][warehouse].update(
                {"db_name": db_name}
            )
        # TODO - Enable support for other warehouses below as developed
        # elif warehouse == "bigquery":
        #     pass

        with open(reflekt_config_path, "w") as f:
            yaml.dump(reflekt_config_obj, f, indent=2)

    click.echo(
        f"Configured to use Reflekt config profile '{config_name}' at "
        f"{reflekt_config_path}"
    )
    project_template_dir = pkg_resources.resource_filename(
        "reflekt", "templates/project/"
    )
    shutil.copytree(project_template_dir, project_dir, dirs_exist_ok=True)

    # Template reflekt_project.yml
    with open(project_yml, "r") as f:
        project_yml_str = f.read()

    project_yml_str = project_yml_str.replace("default_project", project_name).replace(
        "default_profile", config_name
    )

    with open(project_yml, "w") as f:
        f.write(project_yml_str)

    # Template Reflekt project README
    with open(readme_md, "r") as f:
        readme_md_str = f.read()

    readme_md_str = readme_md_str.replace("PROJECT_NAME", project_name).replace(
        "default_profile", config_name
    )

    with open(readme_md, "w") as f:
        f.write(readme_md_str)

    logger.info(
        f"Your Reflekt project '{project_name}' has been created!"
        f"\n\nWith reflekt, you can:\n\n"
        f"    reflekt new --name <plan-name>\n"
        f"        Create a new tracking plan, defined as code.\n\n"  # noqa: E501
        f"    reflekt pull --name <plan-name>\n"
        f"        Get tracking plan from Analytics Governance tool and convert it to code.\n\n"  # noqa: E501
        f"    reflekt push --name <plan-name>\n"
        f"        Sync tracking plan code to Analytics Governance tool. Reflekt handles conversion.\n\n"  # noqa: E501
        f"    reflekt test --name <plan-name>\n"
        f"        Test tracking plan code for naming and metadata conventions (defined in reflect_project.yml).\n\n"  # noqa: E501
        f"    reflekt dbt --name <plan-name>\n"
        f"        Template dbt package with sources, models, and docs for all events in tracking plan."  # noqa: E501
    )

    if plan_type == "avo":
        print("")  # Terminal newline
        logger.info(avo_end_msg)


@click.command()
@click.option(
    "--name",
    "plan_name",
    required=True,
    help="Name of the tracking plan you want to create.",
)
def new(plan_name: str) -> None:
    """Create a new empty tracking plan using provided name."""
    plan_template_dir = pkg_resources.resource_filename("reflekt", "templates/plan/")
    plan_dir = ReflektProject().project_dir / "tracking-plans" / plan_name

    try:
        logger.info(f"Creating new tracking plan '{plan_name}' in '{plan_dir}'")
        shutil.copytree(plan_template_dir, plan_dir)
    except FileExistsError:
        logger.error(
            f"Tracking plan {plan_name} already exists! Please provide a "
            f"different name."
        )
        raise click.Abort()

    plan_yml_file = plan_dir / "plan.yml"

    with open(plan_yml_file, "r") as f:
        doc = yaml.safe_load(f)

    doc["display_name"] = plan_name

    with open(plan_yml_file, "w") as f:
        yaml.dump(doc, f)

    print("")  # Terminal newline
    logger.info(f"[SUCCESS] Created Reflekt tracking plan '{plan_name}'")


@click.command()
@click.option(
    "--name",
    "plan_name",
    required=True,
    help="Tracking plan name in CDP or Analytics Governance tool.",
)
@click.option(
    "--raw",
    flag_value=True,
    help="Pull raw tracking plan JSON (not in Reflekt schema) from CDP.",
)
@click.option(
    "--avo-branch",
    "avo_branch",
    default=None,
    help=("Specify the branch name you want to pull your Avo tracking plan from."),
)
def pull(plan_name: str, raw: bool, avo_branch: str) -> None:
    """Generate tracking plan as code using the Reflekt schema."""
    api = ReflektApiHandler().get_api(avo_branch=avo_branch)
    config = ReflektConfig()
    logger.info(
        f"Searching {titleize(config.plan_type)} for tracking plan '{plan_name}'"
    )
    plan_json = api.get(plan_name)
    logger.info(f"Found tracking plan '{plan_name}'")

    if raw:
        logger.info(
            f"Displaying raw {titleize(config.plan_type)} tracking plan '{plan_name}'\n"
        )
        click.echo(json.dumps(plan_json, indent=2))
    else:
        plan_dir = ReflektProject().project_dir / "tracking-plans" / plan_name

        if api.type.lower() == "avo":
            plan = AvoPlan(plan_json, plan_name)
        elif api.type.lower() == "segment":
            plan = SegmentPlan(plan_json)
        # TODO: Add support for other tracking plan types
        # elif config.plan_type.lower() == "iteratively":
        #     plan = IterativelyPlan(plan_json)
        # elif config.plan_type.lower() == "rudderstack":
        #     plan = RudderstackPlan(plan_json)
        # elif config.plan_type.lower() == "snowplow":
        #     plan = SnowplowPlan(plan_json)

        logger.info(f"Building Reflekt tracking plan '{plan_name}' at {plan_dir}")
        plan.build_reflekt(plan_dir)
        print("")  # Terminal newline
        logger.info(f"[SUCCESS] Built Reflekt tracking plan '{plan_name}'")


@click.command()
@click.option(
    "--name",
    "plan_name",
    required=True,
    help="Tracking plan name in CDP or Analytics Governance tool.",
)
@click.option(
    "--dry",
    is_flag=True,
    help="Output JSON to be synced, without actually syncing it.",
)
def push(plan_name, dry) -> None:
    """Sync tracking plan to CDP or Analytics Governance tool."""
    api = ReflektApiHandler().get_api()
    if api.type.lower() in ["avo", "iteratively"]:
        logger.error(f"'reflekt push' not supported for {titleize(api.type)}.")
        raise click.Abort()

    plan_dir = ReflektProject().project_dir / "tracking-plans" / plan_name
    logger.info(f"Loading Reflekt tracking plan '{plan_name}'")
    loader = ReflektLoader(plan_dir=plan_dir, plan_name=plan_name)
    reflekt_plan = loader.plan
    transformer = ReflektTransformer(reflekt_plan)
    cdp_plan = transformer.build_cdp_plan()

    if dry:
        payload = api.sync(plan_name, cdp_plan, dry=True)
        logger.info(
            f"[DRY RUN] The following JSON would be sent to {transformer.plan_type}"
        )
        click.echo(json.dumps(payload, indent=2))
    else:
        print("")  # Terminal newline
        logger.info(
            f"Syncing converted tracking plan '{plan_name}' to "
            f"{titleize(transformer.plan_type)}"
        )
        api.sync(plan_name, cdp_plan)
        print("")  # Terminal newline
        logger.info(
            f"[SUCCESS] Synced Reflekt tracking plan '{plan_name}' to "
            f"{titleize(transformer.plan_type)}"
        )


@click.command()
@click.option(
    "--name",
    "plan_name",
    required=True,
    help="Tracking plan name in CDP or Analytics Governance tool.",
)
def test(plan_name: str) -> None:
    """Test tracking plan schema for naming, data types, and metadata."""
    plan_dir = ReflektProject().project_dir / "tracking-plans" / plan_name
    logger.info(f"Testing Reflekt tracking plan '{plan_name}'")

    # Initialize ReflektLoader() always runs checks. Simple, but inelegant.
    ReflektLoader(plan_dir=plan_dir, plan_name=plan_name)

    # If no errors are thrown, passed tests
    logger.info("")
    logger.info(f"[PASSED] No errors detected in Reflekt tracking plan '{plan_name}'")


@click.option(
    "--name",
    "plan_name",
    required=True,
    help="Tracking plan name in your Reflekt project.",
)
@click.option(
    "--schema",
    "schema",
    required=False,
    help=(
        "Schema Reflekt will search when looking for raw event tables that match "
        "tracking plan."
    ),
)
@click.option(
    "--force-version",
    "force_version",
    required=False,
    help="Force Reflekt to template the dbt package with a specified semantic version.",
)
@click.command()
def dbt(
    plan_name: str,
    schema: Optional[str] = None,
    force_version: Optional[str] = None,
) -> None:
    """Build dbt package with sources, models, and docs based on tracking plan."""
    project_dir = ReflektProject().project_dir
    plan_dir = project_dir / "tracking-plans" / plan_name
    dbt_pkgs_dir = project_dir / "dbt_packages"
    logger.info(f"Loading Reflekt tracking plan {plan_name}")
    loader = ReflektLoader(plan_dir=plan_dir, plan_name=plan_name, schema_name=schema)
    reflekt_plan = loader.plan
    logger.info(f"Loaded Reflekt tracking plan {plan_name}\n")
    pkg_suffix = (
        reflekt_plan.schema_alias
        if reflekt_plan.schema_alias is not None
        else reflekt_plan.schema
    )
    pkg_name = f"reflekt_{pkg_suffix}"
    dbt_project_yml_path = dbt_pkgs_dir / pkg_name / "dbt_project.yml"

    if force_version:  # If user has forced version, use that
        try:
            version = parse_version(force_version)
        except InvalidVersion:
            logger.error(f"[ERROR] Invalid semantic version provided: {force_version}")
            raise click.Abort()
    # If we don't find a dbt_project.yml -> package does not exist -> v0.1.0
    elif not dbt_project_yml_path.exists():
        version = parse_version("0.1.0")
    else:  # Package already exists!
        with dbt_project_yml_path.open() as f:
            dbt_project_yml_path = yaml.safe_load(f)

        existing_version = parse_version(dbt_project_yml_path["version"])
        new_version_str = (
            f"{existing_version.major}."
            f"{existing_version.minor + 1}."
            f"{existing_version.micro}"
        )
        bumped_version = parse_version(new_version_str)

        if existing_version:
            logger.info(
                f"[WARNING] Existing dbt package found:\n"
                f"    Package name: {pkg_name}\n"
                f"    Existing version: {existing_version}\n"
                f"    Bumped version: {bumped_version}\n"
            )

        bump = click.confirm(
            f"Do you want to bump dbt package '{pkg_name}' to version {bumped_version}?",
            default=True,
        )

        if bump:
            version = bumped_version
        else:
            overwrite = click.confirm(
                f"[WARNING] Reflekt will overwrite current version of dbt package '{pkg_name}'.\n"  # noqa: E501
                f"    Do you want to continue?",
                default=False,
            )

            if not overwrite:
                raise click.Abort()

            print("")  # Newline in terminal
            version = existing_version

    transformer = ReflektTransformer(
        reflekt_plan=reflekt_plan,
        dbt_package_name=pkg_name,
        pkg_version=version,
    )
    transformer.build_dbt_package()

    create_tag = click.confirm(
        f"Would you like to create a Git commit and tag to easily reference this "
        f"Reflekt dbt package {pkg_name} in your dbt project?",
        default=False,
    )

    if create_tag:
        commit_msg = click.prompt(
            "Git commit message will be",
            type=str,
            default=f"build: {pkg_name}",
        )
        tag = click.prompt(
            "Git tag will be",
            type=str,
            default=f"v{str(version)}__{pkg_name}",
        )
        git_executable = shutil.which("git")
        rel_pkg_path = f"dbt_packages/{pkg_name}"
        rel_avo_json_path = ".reflekt/avo/avo.json"
        subprocess.call(
            args=[
                git_executable,
                "add",
                rel_pkg_path,
                rel_avo_json_path,
            ],
            cwd=project_dir,
        )
        subprocess.call(
            args=[
                git_executable,
                "commit",
                "-m",
                commit_msg,
            ],
            cwd=project_dir,
        )
        subprocess.call(
            args=[
                git_executable,
                "tag",
                tag,
            ],
            cwd=project_dir,
        )


# Add CLI commands to CLI group
cli.add_command(init)
cli.add_command(new)
cli.add_command(pull)
cli.add_command(push)
cli.add_command(test)
cli.add_command(dbt)


# Used for CLI debugging
if __name__ == "__main__":
    # Call CLI command here with arguments as a list
    # init(["--project-dir", "~/Repos/test-repo"])
    # new(["--project-dir", "test-plan"])
    # pull(["--name", "my-plan"])
    # push(["--name", "my-plan"])
    # test(["--name", "my-plan"])
    dbt(
        [
            "--name",
            "my-plan",
            "--schema",
            "patty_bar_web",
            "--force-version",
            "0.1.0",
        ]
    )
    # pull(["--name", "tracking-plan-example"])
    # push(["--name", "tracking-plan-example"])
    # test(["--name", "tracking-plan-example"])



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
# SPDX-License-Identifier: Apache-2.0
#
# SPDX-FileCopyrightText: 2021 Buffer
# SPDX-License-Identifier: MIT


# The function parse_segment_property is a derivative work based on the
# function parse_property from project tracking-plan-kit licensed under MIT.
# All changes are licensed under Apache-2.0.
def parse_segment_property(name: str, property_json: dict, required: list = []):
    p_types = property_json.get("type")
    if not isinstance(p_types, (list,)):
        p_types = [p_types]  # If p_types is not list, make it one

    if not p_types[0]:
        p_types[0] = "any"

    p = {
        "name": name,
        "description": property_json.get("description"),
        "type": p_types[0],
    }
    allow_null = len(p_types) > 1 and p_types[1] == "null"

    if allow_null:
        p["allow_null"] = True

    if required is not None and name in required:
        p["required"] = True

    if "enum" in property_json.keys():
        p["enum"] = property_json.get("enum")

    if "pattern" in property_json.keys():
        p["pattern"] = property_json.get("pattern")

    if "format" in property_json.keys():
        if property_json.get("format") == "date-time":
            p["datetime"] = True
        else:
            pass

    # If the property data type is 'array' or 'object', use recursion to get
    # any nested item schema or nested property schema. A simple intro to
    # recursion in Python can be found at:
    # https://realpython.com/python-recursion/
    if p["type"] == "array":
        # If a schema for items in the array has been defined in the
        # tracking plan, get the schema.
        if "items" in property_json.keys():
            array_props = []
            reqd_array_props = property_json.get("items").get("required")

            for property_name, property_prop in sorted(
                property_json.get("items").get("properties").items()
            ):
                item_p = parse_segment_property(
                    property_name, property_prop, reqd_array_props
                )  # Using RECURSION
                array_props.append(item_p)
            p["array_item_schema"] = array_props
        return p
    elif p["type"] == "object":
        # If a schema is defined for the object property, get the schema.
        object_props = []
        reqd_object_props = property_json.get("required")

        if "properties" in property_json.keys():
            for property_name, property_prop in sorted(
                property_json.get("properties").items()
            ):
                obj_p = parse_segment_property(
                    property_name, property_prop, reqd_object_props
                )  # Using RECURSION
                object_props.append(obj_p)
            p["object_properties"] = object_props
        return p
    else:
        return p


# The function parse_segment_event is a derivative work based on the
# function parse_event from project tracking-plan-kit licensed under MIT.
# All changes are licensed under Apache-2.0.
def parse_segment_event(event_json: dict):
    metadata_raw = event_json.get("rules").get("labels")

    metadata = {}

    for key, value in sorted(metadata_raw.items()):
        metadata.update({key: value})

    event_obj = {
        "name": event_json.get("name"),
        "description": event_json.get("description"),
        "version": event_json.get("version"),
        "metadata": metadata,
    }
    properties = (
        event_json.get("rules").get("properties").get("properties").get("properties")
    )
    required = (
        event_json.get("rules").get("properties").get("properties").get("required", [])
    )
    event_obj_properties = []

    if properties is None:
        pass
    else:
        for name, prop in sorted(properties.items()):
            # logger.info(f"   Parsing property: {name}")
            p = parse_segment_property(name, prop, required)
            event_obj_properties.append(p)

    event_obj["properties"] = event_obj_properties

    return event_obj



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
# SPDX-License-Identifier: Apache-2.0
#
# SPDX-FileCopyrightText: 2021 Buffer
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Optional

import yaml
from loguru import logger

from reflekt.plan import ReflektPlan
from reflekt.project import ReflektProject


# The class ReflektLoader is a derivative work based on the class
# PlanLoader from project tracking-plan-kit licensed under MIT. All
# changes are licensed under Apache-2.0.
class ReflektLoader(object):
    def __init__(
        self, plan_dir: Path, plan_name: str, schema_name: Optional[str] = None
    ) -> None:
        # self._validation_errors = []
        if ReflektProject().exists:
            self.plan_name = plan_name
            self.schema_name = schema_name
            self._load_plan_file(plan_dir / "plan.yml")
            self._load_events(plan_dir / "events")
            self._load_identify_traits(plan_dir / "identify-traits.yml")
            self._load_group_traits(plan_dir / "group-traits.yml")
            self.plan.validate_plan()

    def _load_plan_file(self, path: Path) -> None:
        with open(path, "r") as plan_file:
            yaml_obj = yaml.safe_load(plan_file)
            self.plan = ReflektPlan(
                plan_yaml_obj=yaml_obj,
                plan_name=self.plan_name,
                schema_name=self.schema_name,
            )

    def _load_events(self, path: Path) -> None:
        for file in sorted(Path(path).glob("**/*.yml")):
            logger.info(
                f"    Parsing event file {file.name}",
            )

            with open(file, "r") as event_file:
                yaml_event_obj = yaml.safe_load(event_file)
                for event_version in yaml_event_obj:
                    self.plan.add_event(event_version)

    def _load_identify_traits(self, path: Path) -> None:
        if not path.exists():
            return

        with open(path, "r") as identify_file:
            yaml_obj = yaml.safe_load(identify_file)
            for trait in yaml_obj.get("traits", []):
                self.plan.add_identify_trait(trait)

    def _load_group_traits(self, path) -> None:
        if not path.exists():
            return

        with open(path, "r") as group_file:
            yaml_obj = yaml.safe_load(group_file)
            for trait in yaml_obj.get("traits", []):
                self.plan.add_group_trait(trait)



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
import shutil
from pathlib import Path

import yaml
from inflection import dasherize, underscore
from loguru import logger

from reflekt.avo.parser import parse_avo_event
from reflekt.dumper import ReflektYamlDumper


class AvoPlan(object):
    def __init__(self, plan_json: dict, plan_name: str) -> None:
        self.plan_json = plan_json
        self.name = plan_name

    @classmethod
    def parse_string(cls, json_string: str) -> dict:
        parsed_json = json.loads(json_string)
        return cls(parsed_json)

    @classmethod
    def parse_file(cls, json_file_path: Path) -> dict:
        with open(json_file_path, "r") as f:
            contents = f.read()
        return cls.parse_string(contents)

    def build_reflekt(self, plan_dir: Path) -> None:
        events_dir = plan_dir / "events"
        # If plan directory already exists, remove it.
        if plan_dir.is_dir():
            shutil.rmtree(plan_dir)
        # Re-make directories for Reflekt tracking plan(s)
        for dir in [plan_dir, events_dir]:
            if not dir.exists():
                dir.mkdir()

        self._build_reflekt_plan_file(plan_dir)

        for event_json in self.plan_json.get("events", []):
            self._build_reflekt_event_file(events_dir, event_json)

    def _build_reflekt_plan_file(self, plan_dir: Path) -> None:
        plan_file = plan_dir / "plan.yml"
        plan_obj = {"name": self.name}
        with open(plan_file, "w") as f:
            yaml.dump(plan_obj, f)

    def _build_reflekt_event_file(self, events_dir: Path, event_json: dict) -> None:
        event_name = event_json.get("name")
        event_file_name = dasherize(
            underscore(
                event_json["name"]
                .replace("-", "")  # Remove *existing* hyphens in event name
                .replace("  ", " ")  # double spaces --> single spaces
                .replace(" ", "-")  # spaces --> hyphens
                .replace("/", "")  # Remove slashes
            )
        )
        event_file = events_dir / f"{event_file_name}.yml"
        logger.info(f"    Building Reflekt event '{event_name}'")
        event_obj = parse_avo_event(event_json)
        event_obj_with_version = {"version": 1}  # No event versions in Avo. Set to 1
        event_obj_with_version.update(event_obj)

        with open(event_file, "w") as f:
            yaml.dump(
                [event_obj_with_version],
                f,
                indent=2,
                width=70,
                Dumper=ReflektYamlDumper,
                sort_keys=False,
                default_flow_style=False,
                allow_unicode=True,
                encoding=("utf-8"),
            )



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

import yaml

from reflekt.constants import WAREHOUSES
from reflekt.errors import ReflektConfigError
from reflekt.project import ReflektProject


class ReflektConfig:
    def __init__(self, raise_config_errors: bool = True) -> None:
        if ReflektProject().exists:
            try:
                self.config_errors = []
                self.project = ReflektProject()
                self.config_profile = self.project.config_profile

                if self.project.config_path is None:
                    self.path = Path.home() / ".reflekt" / "reflekt_config.yml"
                else:
                    self.path = self.project.config_path

                self.config = self._get_config()
                self.plan_type = self.config.get("plan_type")
                self.avo_json_source = self.config.get("avo_json_source")
                self.cdp_name = self.config.get("cdp")
                self.access_token = self.config.get("access_token")
                self.workspace_name = self.config.get("workspace_name")
                self.warehouse = self.config.get("warehouse")
                self.warehouse_type = self._get_warehouse_type()

            except ReflektConfigError as error:
                if raise_config_errors:
                    raise error
                else:
                    self.config_errors.append(error)

    def _get_config(self) -> dict:
        try:
            with open(self.path) as f:
                config_yml = yaml.safe_load(f)
            return config_yml[self.config_profile]
        except FileNotFoundError:
            raise ReflektConfigError(
                f"\nNo reflekt_config.yml file found at: {self.path}\nPlease create one."
            )

    def _get_warehouse_type(self) -> str:
        if len(self.warehouse.keys()) > 1:
            raise ReflektConfigError(
                f"Multiple warehouses defined for "
                f"`config_profile: {self.config_profile}' in reflekt_config.yml at "
                f"{self.path}"
                f"\nOnly one warehouse can be defined."
            )

        warehouse_type = list(self.warehouse.keys())[0]

        if warehouse_type not in WAREHOUSES:
            raise ReflektConfigError(
                f"Unknown warehouse {warehouse_type} specified in reflekt_config.yml"
                f" at {self.path}."
            )

        return warehouse_type



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
import shutil
import subprocess
from typing import Optional

from loguru import logger

from reflekt.logger import logger_config
from reflekt.config import ReflektConfig
from reflekt.project import ReflektProject


class AvoCli:
    def __init__(self, avo_branch: Optional[str] = None) -> None:
        self._project = ReflektProject()
        self._config = ReflektConfig()
        self.type = self._config.plan_type
        self.avo_json_source = self._config.avo_json_source
        self.avo_dir = self._project.project_dir / ".reflekt" / "avo"
        self.avo_branch = avo_branch
        logger.configure(**logger_config)

    def get(self, plan_name: str) -> None:
        # warehouse_schemas = self._project.warehouse_schemas
        avo_json_file = self.avo_dir / f"{plan_name}.json"

        if not avo_json_file.exists():
            avo_json_file.touch()
            avo_json_file.write_text("{}")

        self._run_avo_pull(plan_name)

        with open(avo_json_file) as f:
            return json.load(f)

    def _run_avo_pull(self, plan_name: str) -> None:
        logger.info(f"Fetching plan {plan_name} from Avo.\n")
        avo_executable = shutil.which("avo")

        # Use the provided --avo-branch argument if one was provided
        if self.avo_branch is not None:
            subprocess.call(
                args=[
                    avo_executable,
                    "pull",
                    plan_name,
                    "--force",  # --force flag used in case object type used in plan
                    "--branch",
                    self.avo_branch,
                ],
                cwd=self.avo_dir,
            )
        else:
            subprocess.call(
                args=[
                    avo_executable,
                    "pull",
                    plan_name,
                    "--force",  # --force flag used in case object type used in plan
                ],
                cwd=self.avo_dir,
            )

        print("")  # Make output look nicer



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Tuple

import sqlalchemy
from snowflake.sqlalchemy import URL as snow_url
from sqlalchemy.engine.url import URL as redshift_url

from reflekt.config import ReflektConfig
from reflekt.errors import ReflektConfigError


class WarehouseConnection:
    """Class that handles connection to the data warehouse specified in
    reflekt_config.yml
    """

    def __init__(self) -> None:
        self._config = ReflektConfig()
        self.warehouse = self._config.warehouse
        self.warehouse_type = self._config.warehouse_type

        if self.warehouse_type == "redshift":
            self.engine = sqlalchemy.create_engine(
                redshift_url.create(
                    drivername="redshift+redshift_connector",
                    host=self.warehouse.get("redshift").get("host_url"),
                    port=self.warehouse.get("redshift").get("port"),
                    database=self.warehouse.get("redshift").get("db_name"),
                    username=self.warehouse.get("redshift").get("user"),
                    password=self.warehouse.get("redshift").get("password"),
                )
            )
        elif self.warehouse_type == "snowflake":
            self.engine = sqlalchemy.create_engine(
                snow_url(
                    account=self.warehouse.get("snowflake").get("account"),
                    user=self.warehouse.get("snowflake").get("user"),
                    password=self.warehouse.get("snowflake").get("password"),
                    role=self.warehouse.get("snowflake").get("role"),
                    database=self.warehouse.get("snowflake").get("database"),
                    warehouse=self.warehouse.get("snowflake").get("warehouse"),
                )
            )

        else:
            raise ReflektConfigError(
                f"Invalid warehouse type specified in {self._config.path}. See "
                f"Reflekt docs on reflekt_config.yml configuration:"
                f"    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-config"  # noqa: E501
            )

    def get_columns(
        self, schema: str, table_name: str
    ) -> Tuple[Optional[list], Optional[str]]:
        with self.engine.connect() as conn:
            try:
                conn.detach()
                columns = (
                    conn.execute(f"select * from {schema}.{table_name} limit 0")
                    .keys()
                    ._keys
                )
                error_msg = None
                return columns, error_msg
            except sqlalchemy.exc.ProgrammingError as e:
                columns = None

                # Handle error according to warehouse type
                if self.warehouse_type == "redshift":
                    error_msg = e.orig.args[0]["M"]
                elif self.warehouse_type == "snowflake":
                    error_msg = e.orig.msg

                return columns, error_msg



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0


class AmplitudeApi:
    # fmt: off
    """API class to pass tracking plan docs and metadata to Amplitude's
    Taxonomy API. In particular, the ability to update Event properties in
    Amplitude to ensure they are consistent with the tracking plan.

    Amplitude docs:
    https://developers.amplitude.com/docs/taxonomy-api#update-an-event-property
    """
    # fmt: on



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
# SPDX-License-Identifier: Apache-2.0
#
# SPDX-FileCopyrightText: 2021 Buffer
# SPDX-License-Identifier: MIT

import json
import shutil
from pathlib import Path

import funcy
import yaml
from inflection import dasherize, underscore
from loguru import logger

from reflekt.dumper import ReflektYamlDumper
from reflekt.segment.parser import parse_segment_event, parse_segment_property


# The class SegmentPlan is a derivative work based on the class
# JsonTrackingPlan from project tracking-plan-kit licensed under MIT. All
# changes are licensed under Apache-2.0.
class SegmentPlan(object):
    def __init__(self, plan_json: dict):
        self.plan_json = plan_json

    @classmethod
    def parse_string(cls, json_string: str):
        parsed_json = json.loads(json_string)
        plan = cls(parsed_json)
        return plan

    @classmethod
    def parse_file(cls, json_file_path: Path):
        with open(json_file_path, "r") as f:
            contents = f.read()
        return cls.parse_string(contents)

    @property
    def display_name(self):
        return self.plan_json.get("display_name")

    @property
    def name(self):
        return self.plan_json.get("name")

    def build_reflekt(self, plan_dir: Path):
        events_dir = plan_dir / "events"
        # Always start from a clean Reflekt plan
        if plan_dir.is_dir():
            shutil.rmtree(plan_dir)
        for dir in [plan_dir, events_dir]:
            if not dir.exists():
                dir.mkdir()

        self._build_reflekt_plan_file(plan_dir)
        traits_json = (
            self.plan_json.get("rules")
            .get("identify", {})
            .get("properties", {})
            .get("traits", {})
            .get("properties")
        )

        if traits_json:
            logger.info("    Writing Reflekt identify traits to identify.yml")
            self._build_reflekt_identify_file(plan_dir, traits_json)

        group_traits_json = (
            self.plan_json.get("rules")
            .get("group", {})
            .get("properties", {})
            .get("traits", {})
            .get("properties")
        )

        if group_traits_json:
            logger.info("   Writing Reflekt group traits to group.yml")
            self._build_reflekt_group_file(plan_dir, group_traits_json)

        for event_json in self.plan_json.get("rules", {}).get("events", []):
            self._build_reflekt_event_file(events_dir, event_json)

    def _build_reflekt_plan_file(self, plan_dir: Path):
        plan_file = plan_dir / "plan.yml"
        plan_obj = {
            "name": self.display_name
        }  # Segment refers to the plan name as the `display_name` in their API

        with open(plan_file, "w") as f:
            yaml.dump(plan_obj, f)

    def _build_reflekt_event_file(self, events_dir: Path, event_json: dict):
        event_name = event_json.get("name")
        event_file_name = dasherize(
            underscore(
                event_json["name"]
                .replace("-", "")  # Remove *existing* hyphens in event name
                .replace("  ", " ")  # double spaces --> single spaces
                .replace(" ", "-")  # spaces --> hyphens
                .replace("/", "")  # Remove slashes
            )
        )
        event_file = events_dir / f"{event_file_name}.yml"
        logger.info(f"    Writing Reflekt event '{event_name}' to {event_file_name}.yml")
        event_obj = parse_segment_event(event_json)
        event_obj_sorted = {"version": event_obj["version"]}
        remainder_dict = funcy.omit(event_obj, "version")
        event_obj_sorted.update(remainder_dict)

        # If event file exists, then we have multiple event versions. Append this version
        if Path(event_file).is_file():
            with open(event_file, "a") as f:
                f.write("\n")
                yaml.dump(
                    [event_obj_sorted],
                    f,
                    indent=2,
                    width=70,
                    Dumper=ReflektYamlDumper,
                    sort_keys=False,
                    default_flow_style=False,
                    allow_unicode=True,
                    encoding=("utf-8"),
                )
        else:
            with open(event_file, "w") as f:
                yaml.dump(
                    [event_obj_sorted],
                    f,
                    indent=2,
                    width=70,
                    Dumper=ReflektYamlDumper,
                    sort_keys=False,
                    default_flow_style=False,
                    allow_unicode=True,
                    encoding=("utf-8"),
                )

    def _build_reflekt_identify_file(self, plan_dir: Path, traits_json: dict):
        traits = [
            parse_segment_property(name, prop_json)
            for (name, prop_json) in sorted(traits_json.items())
        ]
        traits_obj = {"traits": traits}
        traits_file = plan_dir / "identify-traits.yml"

        with open(traits_file, "w") as f:
            yaml.dump(
                traits_obj,
                f,
                indent=2,
                width=70,
                Dumper=ReflektYamlDumper,
                sort_keys=False,
                default_flow_style=False,
                allow_unicode=True,
                encoding=("utf-8"),
            )

    def _build_reflekt_group_file(self, plan_dir: Path, group_traits_json: dict):
        traits = [
            parse_segment_property(name, prop_json)
            for (name, prop_json) in sorted(group_traits_json.items())
        ]
        traits_obj = {"traits": traits}
        traits_file = plan_dir / "group-traits.yml"

        with open(traits_file, "w") as f:
            yaml.dump(
                traits_obj,
                f,
                indent=2,
                width=70,
                Dumper=ReflektYamlDumper,
                sort_keys=False,
                default_flow_style=False,
                allow_unicode=True,
                encoding=("utf-8"),
            )



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

from reflekt.avo.cli import AvoCli
from reflekt.config import ReflektConfig
from reflekt.segment.api import SegmentApi


class ReflektApiHandler:
    """A class that handles which CDP or Analytics Governance API to use when
    pulling or pushing tracking plans using `reflekt pull` and `reflekt push`
    """

    def __init__(self) -> None:
        self._config = ReflektConfig()
        self.type = self._config.plan_type

    def get_api(self, avo_branch: Optional[str] = None):
        if self._config.plan_type == "avo":
            return AvoCli(
                avo_branch
            )  # Actually a CLI, but called via get_api for naming consistency
        elif self._config.plan_type == "iteratively":
            pass
        elif self._config.plan_type == "rudderstack":
            pass
        elif self._config.plan_type == "segment":
            return SegmentApi(
                workspace_name=self._config.workspace_name,
                access_token=self._config.access_token,
            )
        elif self._config.plan_type == "snowplow":
            pass



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
# SPDX-License-Identifier: Apache-2.0
#
# SPDX-FileCopyrightText: 2021 Buffer
# SPDX-License-Identifier: MIT

import re

from cerberus import Validator

from reflekt.casing import CAMEL_CASE_RE, SNAKE_CASE_RE, TITLE_CASE_RE
from reflekt.errors import ReflektValidationError
from reflekt.project import ReflektProject
from reflekt.schema import (
    reflekt_item_schema,
    reflekt_nested_property_schema,
    reflekt_property_schema,
)


# The class ReflektProperty is a derivative work based on the class
# YamlProperty from project tracking-plan-kit licensed under MIT. All
# changes are licensed under Apache-2.0.
class ReflektProperty(object):
    def __init__(self, property_yaml: dict) -> None:
        if ReflektProject().exists:
            self._project = ReflektProject()
            self._property_yaml = property_yaml
            self.name = self._property_yaml.get("name")
            self.description = self._property_yaml.get("description")
            self.type = self._property_yaml.get("type")
            self.required = self._property_yaml.get("required", False)
            self.allow_null = self._property_yaml.get("allow_null", False)
            self.enum = self._property_yaml.get("enum")
            self.datetime = self._property_yaml.get("datetime")
            self.array_item_schema = self._property_yaml.get("array_item_schema")
            self.object_properties = self._property_yaml.get("object_properties")
            self.validate_property()

    def _check_description(self) -> None:
        if self.description == "":
            message = f"Property {self.name} has no description."
            raise ReflektValidationError(message)

    def _check_enum(self) -> None:
        if self.type != "string" and self.enum:
            message = (
                f"Property {self.name} is of type {self.type}. Cannot specify "
                f"`enum:` for {self.type} data types. "
                f"Only for string data types"
            )
            raise ReflektValidationError(message)

    def _check_datetime(self) -> None:
        if self.type != "string" and self.datetime:
            message = (
                f"Property {self.name} is of type {self.type}. Cannot specify "
                f"`datetime:` for {self.type} data types. "
                f"Only for string data types"
            )
            raise ReflektValidationError(message)

    def _check_array_item_schema(self) -> None:
        if self.array_item_schema:
            for array_item in self.array_item_schema:
                validator = Validator(reflekt_item_schema)
                is_valid = validator.validate(array_item, reflekt_item_schema)

                if not is_valid:
                    message = (
                        f"for {array_item['name']} defined in property "
                        f"{self.name} - {validator.errors}"
                    )
                    raise ReflektValidationError(message)

        if self.type != "array" and self.array_item_schema:
            if self.type is None:
                print(self._property_yaml)
            message = (
                f"Property {self.name} is of type {self.type}. Cannot specify "
                f"`array_item_schema:` for {self.type} data types. "
                f"Only for array data types"
            )
            raise ReflektValidationError(message)

    def _check_object_properties(self) -> None:
        if self.object_properties:
            for object_property in self.object_properties:
                validator = Validator(reflekt_nested_property_schema)
                is_valid = validator.validate(
                    object_property, reflekt_nested_property_schema
                )

                if not is_valid:
                    message = (
                        f"object property "
                        f"{object_property['name']} - {validator.errors}. "
                        f"Parent property is {self.name}."
                    )
                    raise ReflektValidationError(message)

        if self.type != "object" and self.object_properties:
            message = (
                f"Property {self.name} is of type {self.type}. Cannot specify "
                f"object_properties for {self.type} data types. "
                f"Only for object data types"
            )
            raise ReflektValidationError(message)

    def _check_property_name_case(self) -> None:
        case_rule = self._project.properties_case

        if case_rule is not None:
            rule_str = f"case: {case_rule.lower()}"
            if case_rule.lower() == "title":
                regex = TITLE_CASE_RE
            elif case_rule.lower() == "snake":
                regex = SNAKE_CASE_RE
            elif case_rule.lower() == "camel":
                regex = CAMEL_CASE_RE

            match = bool(re.match(regex, self.name))
            rule_type = "case:"
            rule_str = case_rule.lower()

            if not match:
                raise ReflektValidationError(
                    f"Property name '{self.name}' does not match naming convention"
                    f" defined by '{rule_type} {rule_str}' in reflekt_project.yml "
                    f"\n\nEither: "
                    f"\n    - Rename property to match config. OR;"
                    f"\n    - Change '{rule_type} {rule_str}' in reflekt_project.yml."
                )

    def _check_property_name_numbers(self) -> None:
        allow_numbers = self._project.properties_allow_numbers
        if not allow_numbers:
            contains_number = any(char.isdigit() for char in self.name)

            if contains_number:
                raise ReflektValidationError(
                    f"Property name '{self.name}' does not match naming convention"
                    f" defined by 'allow_numbers: {str(allow_numbers).lower()}' in reflekt_project.yml "  # noqa: E501
                    f"\n\nEither: "
                    f"\n    - Rename property to match config. OR;"
                    f"\n    - Change 'allow_numbers:' rule for events in reflekt_project.yml."  # noqa: E501
                )

    def validate_property(self) -> None:
        """Validate event property against Reflekt schema."""
        validator = Validator(reflekt_property_schema)
        is_valid = validator.validate(self._property_yaml, reflekt_property_schema)

        if not is_valid:
            message = f"for property '{self.name}' - {validator.errors}"
            raise ReflektValidationError(message)

        self._check_property_name_case()
        self._check_property_name_numbers()
        self._check_description()
        self._check_enum()
        self._check_datetime()
        self._check_array_item_schema()
        self._check_object_properties()



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

import yaml
from git import InvalidGitRepositoryError, Repo

from reflekt.errors import ReflektProjectError


class ReflektProject:
    def __init__(self, raise_project_errors: bool = True) -> None:
        self._project_errors = []
        self.project_dir = self._get_project_root(Path.cwd())
        self.exists = False  # Assume project does not exist by default

        if self.project_dir is not None:
            self.project_yml = self.project_dir / "reflekt_project.yml"
            self.exists = True if self.project_yml.exists() else False
            with open(self.project_yml, "r") as f:
                self.project = yaml.safe_load(f)

            try:
                self.validate_project()
            except ReflektProjectError as err:
                if raise_project_errors:
                    raise err
                else:
                    self._project_errors.append(err)

    def validate_project(self):
        self._get_project_name()
        self._get_config_profile()
        self._get_config_path()
        self._get_events_case()
        self._get_events_allow_numbers()
        self._get_events_reserved()
        self._get_properties_case()
        self._get_properties_allow_numbers()
        self._get_properties_reserved()
        self._get_data_types()
        self._get_warehouse_database_obj()
        self._get_warehouse_schema_obj()
        self._get_expected_metadata_schema()
        self._get_dbt_src_prefix()
        self._get_dbt_model_prefix()
        self._get_dbt_model_materialized()
        self._get_dbt_model_incremental_logic()
        self._get_dbt_docs_prefix()
        self._get_dbt_docs_in_folder()

    def _get_project_root(self, path: Path) -> Path:
        try:
            git_repo = Repo(path, search_parent_directories=True)
            git_root = Path(git_repo.git.rev_parse("--show-toplevel"))
            reflekt_project_list = list(git_root.glob("**/reflekt_project.yml"))

            # This excludes the reflekt_project.yml file in the templates/
            # folder of this repo. Only an issue for development. Not users.
            reflekt_project_list = [
                path
                for path in reflekt_project_list
                if "templates/project/reflekt_project.yml" not in str(path)
            ]

            if len(reflekt_project_list) > 1:
                raise ReflektProjectError(
                    f"\n"
                    f"\nFound a git repo at '{str(git_root)}' with more than one "
                    f"Reflekt project. Only one Reflekt project per repo."
                    f"\n"
                    f"\n{reflekt_project_list}"
                )

            # If no reflekt project found (i.e. before 'reflekt init') then
            # return None
            if reflekt_project_list:
                return reflekt_project_list[0].parents[0]  # project root
            else:
                return  # None

        except InvalidGitRepositoryError:
            raise ReflektProjectError(
                "\n"
                "\nGit repository not detected. Your Reflekt project must be inside a Git repo to function correctly."  # noqa E501
                "\nYou can create a git repo by running 'git init' at the root of your Reflekt project."  # noqa E501
            )

    def _get_project_name(self) -> None:
        try:
            self.name = self.project["name"]
        except KeyError:
            raise ReflektProjectError(
                "\nMust define a project name in reflekt_project.yml. See Reflekt "
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_config_profile(self) -> None:
        try:
            self.config_profile = self.project["config_profile"]
        except KeyError:
            raise ReflektProjectError(
                "\nMust define a config profile in reflekt_project.yml. See Reflekt "
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_config_path(self) -> None:
        config_path_check = self.project.get("config_path")

        if config_path_check is not None:
            self.config_path = Path(config_path_check)

            if not self.config_path.exists():
                raise ReflektProjectError(
                    f"\n\nOptional 'config_path:' {str(self.config_path)} in reflekt_project.yml does not exist!"  # noqa E501
                )

            if not self.config_path.is_absolute():
                raise ReflektProjectError(
                    f"\n\n"
                    f"Optional 'config_path:' {str(self.config_path)} in reflekt_project.yml must be an absolute path!"  # noqa E501
                )
        else:
            self.config_path = None

    def _get_events_case(self) -> None:
        try:
            self.events_case = self.project["tracking_plans"]["events"]["naming"]["case"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define a 'case:' rule for event naming conventions in reflekt_project.yml. See Reflekt "  # noqa: E501
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_events_allow_numbers(self) -> None:
        try:
            self.events_allow_numbers = self.project["tracking_plans"]["events"][
                "naming"
            ]["allow_numbers"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define an 'allow_numbers:' rule for events naming in reflekt_project.yml. See Reflekt "  # noqa: E501
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_events_reserved(self) -> None:
        try:
            self.events_reserved = self.project["tracking_plans"]["events"]["naming"][
                "reserved"
            ]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define a 'reserved:' rule for events naming in reflekt_project.yml. See Reflekt "  # noqa: E501
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_properties_case(self) -> None:
        try:
            self.properties_case = self.project["tracking_plans"]["properties"][
                "naming"
            ]["case"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define a 'case:' rule for properties naming convention in reflekt_project.yml. See Reflekt "  # noqa: E501
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_properties_allow_numbers(self) -> None:
        try:
            self.properties_allow_numbers = self.project["tracking_plans"]["properties"][
                "naming"
            ]["allow_numbers"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define an 'allow_numbers:' rule for properties naming in reflekt_project.yml. See Reflekt "  # noqa: E501
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_properties_reserved(self) -> None:
        try:
            self.properties_reserved = self.project["tracking_plans"]["properties"][
                "naming"
            ]["reserved"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define a 'reserved:' rule for properties naming in reflekt_project.yml. Example:"  # noqa E501
                "\n"
                "\ntracking_plans:"
                "\n  properties:"
                "\n    naming:"
                "\n      reserved: ['my_reserved_property']  # List (can be empty)"  # noqa E501
                "\n"
            )

    def _get_data_types(self) -> None:
        try:
            self.data_types = self.project["tracking_plans"]["properties"]["data_types"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define allowed data types for event properties in reflekt_project.yml. See Reflekt "  # noqa: E501
                "docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_warehouse_database_obj(self) -> None:
        try:
            self.warehouse_database_obj = self.project["tracking_plans"]["warehouse"][
                "database"
            ]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define a database for each tracking plan in "
                "reflekt_project.yml. See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_warehouse_schema_obj(self) -> None:
        try:
            self.warehouse_schema_obj = self.project["tracking_plans"]["warehouse"][
                "schema"
            ]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define warehouse 'schema:' config in reflekt_project.yml. "
                " See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_expected_metadata_schema(self) -> None:
        if (
            self.project.get("tracking_plans").get("events").get("expected_metadata")
            is not None
        ):
            self.expected_metadata_schema = (
                self.project.get("tracking_plans").get("events").get("expected_metadata")
            )
        else:
            self.expected_metadata_schema = None

    def _get_dbt_src_prefix(self) -> None:
        try:
            self.src_prefix = self.project["dbt"]["templater"]["sources"]["prefix"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define 'prefix:' for templated dbt sources in reflekt_project.yml. "  # noqa: E501
                "See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_dbt_model_prefix(self) -> None:
        try:
            self.model_prefix = self.project["dbt"]["templater"]["models"]["prefix"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define 'prefix:' for templated dbt models in reflekt_project.yml. "  # noqa: E501
                "See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_dbt_model_materialized(self) -> None:
        try:
            self.materialized = self.project["dbt"]["templater"]["models"][
                "materialized"
            ].lower()
            if self.materialized not in ["view", "incremental"]:
                raise ReflektProjectError(
                    "\n\nInvalid materialized config in reflekt_project.yml. "
                    "Must be either 'view' or 'incremental'. See Reflekt docs on "
                    "project configuration:"
                    "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
                )
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust set 'materialized:' for templated dbt models in "
                "reflekt_project.yml. See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_dbt_model_incremental_logic(self) -> None:
        if self.materialized == "incremental":
            try:
                self.incremental_logic = self.project["dbt"]["templater"]["models"][
                    "incremental_logic"
                ]
            except KeyError:
                raise ReflektProjectError(
                    "\n\nWhen 'materialized: incremental' in reflekt_project.yml, must define incremental logic for templated dbt models. Example:"  # noqa E501
                    "\n\n"
                    "\ndbt:"
                    "\n  templater:"
                    "\n    models:"
                    "\n      incremental_logic: |"
                    "\n      {%- if is_incremental() %}"
                    "\n          where event_timestamp >= (select max(event_timestamp)::date from {{ this }})"  # noqa E501
                    "\n      {%- endif %}"
                    "\n\n See Reflekt docs on project configuration:"
                    "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
                )
        else:
            self.incremental_logic = None

    def _get_dbt_docs_prefix(self) -> None:
        try:
            self.docs_prefix = self.project["dbt"]["templater"]["docs"]["prefix"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define 'prefix:' for templated dbt docs in "
                "reflekt_project.yml. See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )

    def _get_dbt_docs_in_folder(self) -> None:
        try:
            self.docs_in_folder = self.project["dbt"]["templater"]["docs"]["in_folder"]
        except KeyError:
            raise ReflektProjectError(
                "\n\nMust define 'in_folder:' for templated dbt docs in "
                "reflekt_project.yml. See Reflekt docs on project configuration:"
                "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
            )



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

# flake8: noqa
# fmt: off
dbt_src_schema = {
    "version": 2,
    "sources": [
        {
            "name": "",
            "schema": "",
            "database": "",
            "description": "",
            "tables": []
        }
    ]
}

dbt_table_schema = {
    "name": "",
    "description": "",
    # "columns": [],
}

dbt_doc_schema = {
    "version": 2,
    "models": [],
}

dbt_model_schema = {
    "name": "",
    "description": "",
    "columns": [],
}

dbt_column_schema = {
    "name": "",
    "description": "",
}
# fmt: on



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
from typing import Optional

import requests
from requests import Response

from reflekt.config import ReflektConfig
from reflekt.segment.errors import SegmentApiError


class SegmentApi:
    def __init__(self, workspace_name: str, access_token: str):
        self._config = ReflektConfig()
        self.type = self._config.plan_type
        self.workspace_name = workspace_name
        self.access_token = access_token
        self.base_url = (
            f"https://platform.segmentapis.com/v1beta/workspaces/"
            f"{workspace_name}/tracking-plans/"
        )

    def _list_plans(self):
        url, headers = self._setup_url_headers()
        r = requests.get(url=url, headers=headers)
        self._handle_response(r)
        plans = r.json().get("tracking_plans")
        name_to_id = {
            plan.get("display_name"): plan.get("name").split("/")[-1] for plan in plans
        }
        id_to_name = {
            plan.get("name").split("/")[-1]: plan.get("display_name") for plan in plans
        }

        return name_to_id, id_to_name

    def _get_plan_id_from_name(self, plan_name: str, name_to_id_dict: dict):
        try:
            return name_to_id_dict[plan_name]
        except KeyError:
            return None

    def _setup_url_headers(self, plan_id: Optional[str] = None):
        if plan_id is None:
            url = self.base_url
        else:
            url = self.base_url + plan_id

        headers = {
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
        }

        return url, headers

    def _handle_response(self, response: Response):
        response_dict = json.loads(response.text.encode("utf8"))
        if "error" in response_dict:
            raise SegmentApiError(
                f"\n\nSegment Config API error encountered during sync. "
                f"\n    Error type: {response_dict['error']}"
                f"\n    Error code: {response_dict['code']}"
                f"\n    See https://segment.com/docs/config-api/api-design/#errors "  # noqa: E501
                f"for error descriptions."
            )

    def get(self, plan_name: str):
        name_to_id, _ = self._list_plans()
        plan_id = self._get_plan_id_from_name(plan_name, name_to_id)

        if plan_id is None:
            raise SegmentApiError(f"Tracking plan '{plan_name}' was not found.")
        else:
            url, headers = self._setup_url_headers(plan_id=plan_id)

        url, headers = self._setup_url_headers(plan_id=plan_id)
        r = requests.get(url=url, headers=headers)
        self._handle_response(r)

        return r.json()

    def sync(self, plan_name: str, json_plan: dict, dry: bool = False):
        name_to_id, _ = self._list_plans()
        plan_id = self._get_plan_id_from_name(plan_name, name_to_id)
        payload = json.dumps(json_plan)

        if plan_id is None:
            if dry:
                return json_plan
            else:
                url, headers = self._setup_url_headers()
                r = requests.post(url=url, headers=headers, data=payload)
                self._handle_response(r)
        else:
            if dry:
                return json_plan
            else:
                url, headers = self._setup_url_headers(plan_id)
                r = requests.put(url=url, headers=headers, data=payload)
                self._handle_response(r)



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
# SPDX-License-Identifier: Apache-2.0
#
# SPDX-FileCopyrightText: 2021 Buffer
# SPDX-License-Identifier: MIT

from typing import Counter, List, Optional

from cerberus import Validator

from reflekt.config import ReflektConfig
from reflekt.errors import ReflektValidationError
from reflekt.event import ReflektEvent
from reflekt.project import ReflektProject
from reflekt.property import ReflektProperty
from reflekt.schema import reflekt_plan_schema
from reflekt.trait import ReflektTrait


# The class ReflektPlan is a derivative work based on the class
# YamlTrackingPlan from project tracking-plan-kit licensed under MIT. All
# changes are licensed under Apache-2.0.
class ReflektPlan(object):
    def __init__(
        self,
        plan_yaml_obj: dict,
        plan_name: str,
        schema_name: Optional[str] = None,
    ) -> None:
        if ReflektProject().exists:
            self._config = ReflektConfig()
            self._project = ReflektProject()
            self.plan_yaml_obj = plan_yaml_obj
            self.name = plan_name

            if schema_name:
                self.schema_arg = schema_name
                self.database = self._get_database(self.name)
                self.schema, self.schema_alias = self._get_schema_and_schema_alias(
                    self.name
                )
            else:
                self.database = None
                self.schema = None
                self.schema_alias = None

            self.events: List[ReflektEvent] = []
            self.identify_traits: List[ReflektTrait] = []
            self.group_traits: List[ReflektTrait] = []

    def _get_database(self, plan_name: str) -> list:
        try:
            plan_database = self._project.warehouse_database_obj[plan_name]
            return plan_database
        except KeyError:
            raise KeyError(
                f"Tracking plan '{plan_name}' not found under...\n\n"
                f"  tracking_plans:\n"
                f"    warehouse:\n"
                f"      database:\n\n"
                f"... in reflekt_project.yml. See Reflekt docs on "
                f"reflekt_project.yml configuration: "
                f"\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#project-configuration"  # noqa: E501
            )

    def _get_schema_and_schema_alias(self, plan_name: str) -> list:
        try:
            schema_configs = self._project.warehouse_schema_obj[plan_name]

            for schema_config in schema_configs:
                if schema_config["schema"] == self.schema_arg:
                    schema = schema_config["schema"]
                    schema_alias = schema_config.get("schema_alias")

                    return schema, schema_alias

        except KeyError:
            raise KeyError(
                f"Error in 'schema:' config for tracking plan '{plan_name}'."
                f"See Reflekt docs for guidance on 'schema:' config.\n"
                f"    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#project-configuration"  # noqa: E501
            )

    def add_event(self, event_yaml_obj: dict) -> None:
        self.events.append(ReflektEvent(event_yaml_obj))

    def add_identify_trait(self, trait_yaml: dict) -> None:
        trait_property = ReflektProperty(trait_yaml)
        self.identify_traits.append(trait_property)

    def add_group_trait(self, trait_yaml: dict) -> None:
        trait_property = ReflektProperty(trait_yaml)
        self.group_traits.append(trait_property)

    def _check_duplicate_events(self) -> None:
        event_ids = map(lambda e: e.name + str(e.version), self.events)
        counts = Counter(event_ids)
        duplicates = {k: v for (k, v) in counts.items() if v > 1}

        if len(duplicates) > 0:
            duplicate_names = ", ".join(duplicates.keys())
            raise ReflektValidationError(
                f"Duplicate events found. Events: {duplicate_names}"
            )

    def _check_reserved_event_names(self) -> None:
        if len(self.events) == 0:
            return

        event_names = [e.name for e in self.events]

        for event_name in event_names:
            if event_name in ReflektProject().events_reserved:
                raise ReflektValidationError(
                    f"Event name '{event_name}' is reserved and cannot be used. "
                    "See Reflekt docs on project configuration:"
                    "\n    https://github.com/GClunies/reflekt/blob/main/docs/DOCUMENTATION.md#reflekt-project"  # noqa: E501
                )

    def validate_plan(self) -> None:
        validator = Validator(reflekt_plan_schema)
        is_valid = validator.validate(self.plan_yaml_obj, reflekt_plan_schema)

        if not is_valid:
            raise ReflektValidationError(
                f"Invalid plan definition for plan '{self.name}' - {validator.errors}"
            )

        self._check_duplicate_events()
        self._check_reserved_event_names()



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import sys

from reflekt.project import ReflektProject

if ReflektProject().exists:
    sink = (
        str(ReflektProject().project_dir)
        + "/.reflekt/logs/reflekt_{time:YYYY-MM-DD-HHmmss}.log"
    )
else:
    sink = "./.reflekt/logs/reflekt_{time:YYYY-MM-DD-HHmmss}.log"


logger_config = {
    "handlers": [
        {
            "sink": sys.stdout,
            "format": "{time:HH:mm:ss} {message}",
        },
        {
            "sink": sink,
            "serialize": True,
        },
    ],
}



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

TITLE_CASE_RE = r"[A-Z][a-z]+"
CAMEL_CASE_RE = r"(^[a-z]|[A-Z])[a-z]*"
SNAKE_CASE_RE = r"^[a-z._]+$"



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

import yaml


class ReflektYamlDumper(yaml.Dumper):
    """Custom YAML dumper class. Indents list object for readability."""

    def increase_indent(self, flow: bool = False, indentless: bool = False):
        return super(ReflektYamlDumper, self).increase_indent(flow, False)

    # Don't use YAML anchors/aliases
    # https://ttl255.com/yaml-anchors-and-aliases-and-how-to-disable-them/
    def ignore_aliases(self, data):
        return True



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0

from reflekt.project import ReflektProject

# fmt: off
reflekt_plan_schema = {
    "name": {"required": True, "type": "string"},
    "dbt_pkg_schema": {"required": False, "type": "string"},
}

reflekt_event_schema = {
    "version": {"required": True, "type": "integer", "min": 1},
    "name": {"required": True, "type": "string"},
    "description": {"required": True, "type": "string"},
    "metadata": {"required": False, "type": "dict"},
    "properties": {"required": True, "type": "list"},
}

reflekt_project = ReflektProject()

if reflekt_project.exists:
    reflekt_property_schema = {
        "name": {"required": True, "type": "string"},
        "description": {"required": True, "type": "string"},
        "type": {
            "required": True,
            "allowed": reflekt_project.data_types,
        },
        "allow_null": {"required": False, "type": "boolean"},
        "required": {"required": False, "type": "boolean"},
        "enum": {"required": False, "type": "list"},
        "pattern": {"required": False, "type": "string"},
        "datetime": {"required": False, "type": "boolean"},
        "object_properties": {"required": False, "type": "list"},
        "array_item_schema": {"required": False, "type": "list"},
    }

    reflekt_item_schema = {
        "name": {"required": True, "type": "string"},
        "description": {"required": True, "type": "string"},
        "type": {
            "required": True,
            "allowed": reflekt_project.data_types,
        },
        "required": {"required": False, "type": "boolean"},
        "enum": {"required": False, "type": "list"},
        "pattern": {"required": False, "type": "string"},
        "datetime": {"required": False, "type": "boolean"},
    }

    reflekt_nested_property_schema = {
        "name": {"required": True, "type": "string"},
        "description": {"required": True, "type": "string"},
        "type": {
            "required": True,
            "allowed": reflekt_project.data_types,
        },
        "required": {"required": False, "type": "boolean"},
        "enum": {"required": False, "type": "list"},
        "pattern": {"required": False, "type": "string"},
        "datetime": {"required": False, "type": "boolean"},
    }

    if reflekt_project.expected_metadata_schema is not None:
        reflekt_expected_metadata_schema = reflekt_project.expected_metadata_schema

    else:
        reflekt_expected_metadata_schema = None

else:  # These are just dummy objects to prevent import side-effects
    reflekt_expected_metadata_schema = {}
    reflekt_property_schema = {}
    reflekt_item_schema = {}
    reflekt_nested_property_schema = {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reflekt',
 'reflekt.amplitude',
 'reflekt.avo',
 'reflekt.iteratively',
 'reflekt.rudderstack',
 'reflekt.segment',
 'reflekt.snowplow']

package_data = \
{'': ['*'],
 'reflekt': ['templates/*',
             'templates/dbt/*',
             'templates/dbt/analyses/*',
             'templates/dbt/macros/*',
             'templates/dbt/models/*',
             'templates/dbt/seeds/*',
             'templates/dbt/snapshots/*',
             'templates/dbt/tests/*',
             'templates/plan/*',
             'templates/plan/events/*',
             'templates/project/*',
             'templates/project/.reflekt/avo/*',
             'templates/project/.reflekt/logs/*',
             'templates/project/.reflekt/tmp/*',
             'templates/project/dbt_packages/*',
             'templates/project/tracking-plans/example-plan/*',
             'templates/project/tracking-plans/example-plan/events/*']}

install_requires = \
['Cerberus>=1.3.4,<2.0.0',
 'GitPython>=3.1.27,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'click>=8.1.2,<9.0.0',
 'funcy>=1.17,<2.0',
 'inflection>=0.5.1,<0.6.0',
 'loguru>=0.6.0,<0.7.0',
 'redshift-connector>=2.0.905,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'snowflake-sqlalchemy>=1.3.3,<2.0.0',
 'sqlalchemy-redshift>=0.8.9,<0.9.0']

entry_points = \
{'console_scripts': ['reflekt = reflekt:cli']}

setup_kwargs = {
    'name': 'reflekt',
    'version': '0.1.10',
    'description': 'Reflekt lets data teams: 1) Define tracking plans as code; 2) Automagically build a dbt package that models and documents the events in a tracking plan pulled from an Analytics Governance tool, ready for use in a dbt project.',
    'long_description': None,
    'author': 'Greg Clunies',
    'author_email': 'greg.clunies@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0
class ReflektValidationError(Exception):
    pass


class ReflektProjectError(Exception):
    pass


class ReflektConfigError(Exception):
    pass



# SPDX-FileCopyrightText: 2022 Gregory Clunies <greg@reflekt-ci.com>
#
# SPDX-License-Identifier: Apache-2.0


def parse_avo_property(name: str, property_json: dict, required: list = []) -> dict:
    p_type = property_json.get("type")
    p = {
        "name": name,
        "description": property_json.get("description"),
        "type": p_type,
    }

    if required is not None and name in required:
        p["required"] = True

    if "enum" in property_json.keys():
        p["enum"] = property_json.get("enum")

    if p_type == "array":
        # Unlike Segment, Avo takes a simplified approach to array types, user
        # can only specify the type of the array elements. No 'array items'
        # like Segment. This placeholder is kept in case this changes.
        return p

    elif p["type"] == "object":
        object_props = []
        reqd_object_props = property_json.get("required")

        if "properties" in property_json.keys():
            for property_name, property_prop in property_json.get("properties").items():
                obj_p = parse_avo_property(
                    property_name, property_prop, reqd_object_props
                )  # Using RECURSION
                object_props.append(obj_p)
            p["object_properties"] = object_props

        return p

    else:
        return p


def parse_avo_event(event_json: dict) -> dict:
    tags = event_json.get("tags")
    metadata_raw = {
        tag.split(":")[0]: tag.split(":")[1].replace(" ", "") for tag in tags
    }
    metadata = {}

    for key, value in sorted(metadata_raw.items()):
        metadata.update({key: value})

    event_obj = {
        "name": event_json.get("name"),
        "description": event_json.get("description"),
        "metadata": metadata,
    }
    properties_lvl2 = event_json.get("rules").get("properties").get("properties")

    if properties_lvl2 is not None:
        properties = properties_lvl2.get("properties")
        required = properties_lvl2.get("required", [])
    else:
        properties = None
        required = []

    event_obj_properties = []

    if properties is None:
        pass
    else:
        for name, prop in sorted(properties.items()):
            # logger.info(f"Dumping property: {name}")
            p = parse_avo_property(name, prop, required)
            event_obj_properties.append(p)

    event_obj["properties"] = event_obj_properties

    return event_obj
